import os.path
from typing import List
import numpy as np
from pandas import DataFrame
import urllib.request
import tarfile
from numba import njit


@njit()
def dist(point0: np.ndarray, point1: np.ndarray):
    """
    calculates euclidian distance between 2 coordinates in a 2d space from an array
    exists because scipy.spatial.distance.euclidian is not elementwise
    :param point0: the first x,y coordinate
    :param point1: the second x,y coordinate
    :return:
    """
    return np.sqrt(
        np.power(np.abs(point1[:, 0] - point0[:, 0]), 2)
        + np.power(np.abs(point1[:, 1] - point0[:, 1]), 2)
    )


@njit
def scalar_dist(point0: np.ndarray, point1: np.ndarray):
    """
    calculates euclidian distance between 2 coordinates in a 2d space
    exists because scipy.spatial.distance.euclidian is not elementwise
    :param point0: the first x,y coordinate
    :param point1: the second x,y coordinate
    :return:
    """
    return np.sqrt(
        np.power(np.abs(point1[0] - point0[0]), 2) + np.power(np.abs(point1[1] - point0[1]), 2)
    )


def point_array(data_frame: DataFrame, points: List[str], likelihood: bool = False):
    """
    returns a numpy array of the XY coordinates for the Listed points
    :param data_frame: the pandas dataframe from the h5 returned from deeplabcut's analysis
    :param points: a list of the names of the points to return coordinates for
    :param likelihood: if this is set True the likelihood value for each point will be included on the returned array
    """
    nd_arrays = list()
    if likelihood:
        for point in points:
            nd_arrays.append(data_frame[point][["x", "y", "likelihood"]].to_numpy(dtype=np.float_))
    else:
        for point in points:
            nd_arrays.append(data_frame[point][["x", "y"]].to_numpy(dtype=np.float_))
    return np.squeeze(np.stack(nd_arrays, axis=0))


def angle_between_lines(m1: float, m2: float) -> float:
    """
    calculates the angle between 2 lines in degrees
    :param m1: slope of line 1
    :param m2: slope of line 2
    :return: angle
    """
    return float((np.arctan(((m2 - m1) / (1 + m1 * m2)))))


def download_model(url: str, dest: str) -> str:
    """
    downloads a tarball model from url to dest/temp extracts it to dest/tar_name and returns the path of the
    model directory
    :param url: the url to download the model from
    :param dest: the destination to put the model
    :return: the path of the model directory
    """
    temp = os.path.join(dest, "temp")
    if not os.path.isdir(temp):
        os.mkdir(temp)
    model_tar = urllib.request.urlretrieve(url, temp)[0]
    if tarfile.is_tarfile(model_tar):
        ext = os.path.splitext(model_tar)[1]
        if ext == "gz" or ext == ".gz":
            tar = tarfile.open(model_tar, "r:gz")
        elif ext == "xz" or ext == ".xz":
            tar = tarfile.open(model_tar, "r:gz")
        else:
            raise ValueError("File must be .tar.gz or tar.xz")
    else:
        raise ValueError("File must be .tar.gz or tar.xz")
    name = os.path.splitext(os.path.split(model_tar)[1])[0]
    if not os.path.isdir(name):
        os.mkdir(name)
    tar.extractall(name)
    return name


def nan(shape, dtype: type = np.float_, order="C"):
    arr = np.zeros(shape=shape, dtype=dtype, order=order)
    arr[:] = np.nan
    return arr



import os
import sys

try:
    cuda_bin = os.path.join(os.environ.get("CUDA_PATH_V11_2"), "bin")
    os.add_dll_directory(cuda_bin)
except TypeError as e:
    pass
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
from logging import info
import matplotlib as mpl

try:
    import qtpy
except (ModuleNotFoundError, ImportError) as e:
    info("aemotrics gui not loaded")
if "qtpy" in sys.modules and ("PySide2" in sys.modules or "PyQt5" in sys.modules):
    from dlc_generic_analysis import gui_objects, gui_utils
    from dlc_generic_analysis.trimmer import Trimmer
    from dlc_generic_analysis.dlc_generic_analysis import MainWidget
    from dlc_generic_analysis.viewer import ViewerWidget

    mpl.use("QtAgg")
else:
    info("dlc-generic-analysis GUI not loaded")
    mpl.use("agg")
from dlc_generic_analysis import geometries, math_utils, utils, video_tools, filter
from dlc_generic_analysis.dlc import dlc_analyze
from dlc_generic_analysis.analysis import Analysis
from dlc_generic_analysis import line



from typing import Tuple
import numpy as np


class Line:
    """A line between two points. Used to measure scarring or other deformation
    in vocal cords.
    end1: point
        first point to calc line Endpoint with smaller x value if x1 \neq x2
    end2: point
        second point to calc line. Endpoint with greater x if x1 \neq x2
    slope: float
        slope of the line
    yint: float
        y intercept of line
    """

    def __init__(
        self,
        end1: Tuple[float, float] = np.nan,
        end2: Tuple[float, float] = np.nan,
        slope: float = np.nan,
        intercept: float = np.nan,
    ):
        """
        creates a Line either between 2 points or a slope and intercept
        :param end1: The first point for the line.
        :param end2: The second point for the line.
        :param slope: The slope of the line.
        :param intercept: the Y intercept of the line
        """
        if (
            not np.isnan(end1).any()
            and end1 is not None
            and not np.isnan(end2).any()
            and end2 is not None
        ):
            if end1[0] > end2[0]:
                self._end1 = end2
                self._end2 = end1
            else:
                self._end1 = end1
                self._end2 = end2
            self._slope = None
            self._intercept = None
            self._calc_slope_intercept()
        elif (
            not np.isnan(slope)
            and slope is not None
            and not np.isnan(intercept)
            and intercept is not None
        ):
            self._intercept = intercept
            self._slope = slope
            self._end1 = (0, self._intercept)
            self._end2 = (1, int(self._intercept + self._slope))
        else:
            raise AttributeError

    def _calc_slope_intercept(self):
        if not self._end2[0] == self._end1[0]:
            self._slope = (self._end2[1] - self._end1[1]) / (self._end2[0] - self._end1[0])
        else:
            self._slope = 999999999999999
        self._intercept = self._end1[1] - self.slope * self._end1[0]

    @property
    def end1(self):
        return self._end1

    @property
    def end2(self):
        return self._end2

    @end1.setter
    def end1(self, end1: (float, float)):
        self._end1 = end1
        self._calc_slope_intercept()

    @end2.setter
    def end2(self, end2: (float, float)):
        self._end2 = end2
        self._calc_slope_intercept()

    @property
    def slope(self):
        return self._slope

    @property
    def intercept(self):
        return self._intercept



import os.path
import numpy as np
from typing import List
import pandas
from pandas import DataFrame
from logging import info


def threshold_confidence(h5s: List[str], min_likelihood: float, save_csv=False) -> List[str]:
    """
    Thresholds the dlc analysis for a minimum likelihood values less than min_likelihood will be changed to np.nan
    :param h5s: the paths to the h5 files
    :param min_likelihood: the likelihood cutoff all points
    :param save_csv: set true to save a csv file in addition to the hdf5
    :return: a list to the path of thresholded hdf5s
    """
    ret_h5s = []
    for h5_path in h5s:
        if not os.path.isfile(h5_path):
            raise FileNotFoundError(f"{h5_path} is not a file")
        out_path = os.path.splitext(h5_path)[0] + "_threshold.h5"
        ret_h5s.append(out_path)
        if not os.path.isfile(out_path):
            df = pandas.read_hdf(h5_path)
            assert type(df) == DataFrame
            dlc_scorer = df.columns.levels[0].to_list()[0]
            bps = df.columns.levels[1].to_list()
            __threshold_bp(df, dlc_scorer, min_likelihood, bps)
            df.to_hdf(out_path, key="scorer", mode="w")
            if save_csv:
                df.to_csv(out_path.replace(".h5", ".csv"), mode="w")
        else:
            info(f"{h5_path} was already thresholded, skipping")
    return ret_h5s


def __threshold_bp(df: DataFrame, dlc_scorer, cutoff: float, bps: List[str]):
    for i, bp in enumerate(bps):
        prob = df[dlc_scorer, bp, "likelihood"].values.squeeze()
        mask_ = prob < cutoff
        x = np.ma.array(df[dlc_scorer, bp, "x"].values, mask=mask_).filled(np.nan).squeeze()
        y = np.ma.array(df[dlc_scorer, bp, "y"].values, mask=mask_).filled(np.nan).squeeze()
        df[dlc_scorer, bp, "x"] = x
        df[dlc_scorer, bp, "y"] = y



import pandas
import numpy as np

_index = pandas.MultiIndex.from_product(
    [["dlc_sc"], ["Glab", "Ment"], ["x", "y", "likelihood"]],
    names=["scorer", "bodyparts", "coords"],
)

_data = np.array([[0, 1, 0.5, 3, 4, 0.99], [6, 7, 0.8, 1, 2, 0.6], [4, 5, 0.9, 7, 8, 0.9]])
df = pandas.DataFrame(_data, columns=_index)



import os
import moviepy.editor


def resize(path: str, width):
    clip = moviepy.editor.VideoFileClip(path)
    clip_resized = clip.resize(width=width)
    filename, extension = os.path.splitext(path)
    clip_resized.write_videofile(filename + "_resized" + extension)



from qtpy import QtWidgets
import qtawesome as qta


class PlayPause(QtWidgets.QHBoxLayout):
    def __init__(self):
        """
        Creates fast reverse play/pause and fast forward buttons wrapped in a QHBoxLayout
        """
        QtWidgets.QHBoxLayout.__init__(self)
        self.setContentsMargins(0, 0, 0, 0)
        self.fast_reverse_button = QtWidgets.QPushButton()
        fr_icon = qta.icon("fa5s.fast-backward")
        self.fast_reverse_button.setIcon(fr_icon)
        self.addWidget(self.fast_reverse_button)
        self.play_icon = qta.icon("fa5s.play")
        self.pause_icon = qta.icon("fa5s.pause")
        self.play_pause_button = QtWidgets.QPushButton()
        self.play_pause_button.setIcon(self.play_icon)
        self.addWidget(self.play_pause_button)
        ff_icon = qta.icon("fa5s.fast-forward")
        self.fast_forward_button = QtWidgets.QPushButton()
        self.fast_forward_button.setIcon(ff_icon)
        self.addWidget(self.fast_forward_button)



from dlc_generic_analysis import geometries


def test_line_from_mx_b():
    line0 = geometries.Line(slope=2, intercept=0)
    assert line0._end1 == (0, 0)
    assert line0._end2 == (1, 2)


def test_line_from_points():
    line1 = geometries.Line((0, 5), (10, 10))
    assert line1.intercept == 5
    assert line1.slope == 0.5


def test_line_setter():
    line1 = geometries.Line((0, 0), (1, 1))
    line1.end2 = (10, 15)
    line1.end1 = (0, 10)
    assert line1.intercept == 10
    assert line1.slope == 0.5



from dlc_generic_analysis import utils
import numpy as np
import pandas
from numpy import testing


def test_scalar_dist():
    dist = utils.scalar_dist(np.array([10, 2]), np.array([2, 5]))
    np.testing.assert_array_almost_equal(8.54400374531753, dist)
    dist = utils.scalar_dist(np.array([10, 2]), np.array([2, 2]))
    np.testing.assert_array_equal(8, dist)
    dist = utils.scalar_dist(np.array([5, 3]), np.array([8, -2]))
    testing.assert_array_almost_equal(5.830951894845301, dist)


def test_dist():
    dist = utils.dist(np.array([[0, 1], [0, 2]]), np.array([[10, 10], [10, 25]]))
    testing.assert_array_almost_equal(
        np.array([13.45362404707371, 25.079872407968907], dtype=np.float_), dist
    )


def test_point_array():
    data = [[0, 1, 2], [3, 4, 5]]
    index = pandas.MultiIndex.from_product(
        [["Glab"], ["x", "y", "likelihood"]], names=["scorer", "coords"]
    )
    df = pandas.DataFrame(data, columns=index)
    print(df)
    arr = utils.point_array(df, ["Glab"])
    testing.assert_array_equal(arr, np.array([[0, 1], [3, 4]]))


def test_angle_between_lines():
    assert np.round(np.degrees(utils.angle_between_lines(np.sqrt(3), 0)), 6) == -60


def test_nan():
    nans = np.zeros((10, 10, 10), dtype=float)
    nans[:] = np.nan
    testing.assert_array_equal(utils.nan((10, 10, 10), dtype=float), nans)



import numpy as np
from dlc_generic_analysis import math_utils


def test_interpolate_gaps():
    values = [1, 3, 5, 7, 9]
    filled = math_utils.interpolate_gaps(values, 5)
    print(filled)
    assert np.array_equal(np.array([1, 3, 5, 7, 9]), filled)



import numpy as np


def interpolate_gaps(data, limit: int = None) -> np.ndarray:
    """
    interpolates gaps in data series with np.nan
    :param data: the data to interpolate
    :param limit: the maximum gap size to fill
    :return: the fi;;ed data
    """
    data = np.array(data)
    i = np.arange(data.size)
    valid = np.isfinite(data)
    filled = np.interp(i, i[valid], data[valid])

    if limit is not None:
        invalid = ~valid
        for n in range(1, limit + 1):
            invalid[:-n] &= invalid[n:]
        filled[invalid] = np.nan

    return filled



from dlc_generic_analysis.filter import threshold_confidence
import pandas
import numpy as np
from .test_data import df


def test_threshold_confidence():
    path = "test.h5"
    df.to_hdf(path, "dlc_sc")
    path_thresh = threshold_confidence([path], 0.85)[0]
    df_thresh = pandas.read_hdf(path_thresh)
    data_thresh = df_thresh["dlc_sc"].to_numpy()
    np.testing.assert_equal(
        data_thresh,
        [
            [np.nan, np.nan, 0.5, 3, 4, 0.99],
            [np.nan, np.nan, 0.8, np.nan, np.nan, 0.6],
            [4, 5, 0.9, 7, 8, 0.9],
        ],
    )



from dlc_generic_analysis.gui_objects import PlayPause
from qtpy import QtWidgets


# def test_play_pause(qtbot):
#     p_p = PlayPause()
#     w = QtWidgets.QWidget()
#     w.setLayout(p_p)
#     qtbot.addWidget(w)



import logging
import multiprocessing
import os
from qtpy import QtWidgets, QtCore, QtMultimedia, QtMultimediaWidgets
from qtpy.QtCore import Qt
from .gui_objects import PlayPause
from typing import List, Tuple
from moviepy.editor import VideoFileClip


def _trim_pool(args: Tuple[str, float, float]) -> None:
    """
    trims a file and writes it to a new file with "_trimmed" appended to the end
    :param args: A tuple of (str, float, float) where str is the video path, the first float is the
    start of the trim region and the second float is the end of the trim region
    :return: None
    """
    clip = VideoFileClip(args[0])
    clip.cutout(args[1] / 1000, args[2] / 1000)
    fname, ext = os.path.splitext(args[0])
    outfile = fname + "_trimmed" + ext
    clip.write_videofile(outfile)
    return outfile


class GoToTime(QtWidgets.QHBoxLayout):
    def __init__(self, text: str = "", time: float = 0):
        """
        A button that says go to with an editable label
        :param text: the text of the label
        :param time: the initial time value
        """
        super(GoToTime, self).__init__()
        self.text = text
        self.setContentsMargins(0, 0, 0, 0)
        if text != "":
            self.label = QtWidgets.QLabel(f"{text} {round(time, 2)}")
        else:
            self.label = QtWidgets.QLabel("")
        self.addWidget(self.label)
        self.goto_button = QtWidgets.QPushButton("Go To")
        self.addWidget(self.goto_button)

    def update_label(self, time):
        self.label.setText(f"{self.text} {round(time, 2)}")


class Trimmer(QtWidgets.QWidget):
    keyPressed = QtCore.Signal(int)

    def __init__(self, video_paths: List[str]):
        """
        Creates a window that assists the user in trimming the ends off a list of videos
        """
        super(Trimmer, self).__init__()
        self.setWindowTitle("Trim Video(s)")
        self.setLayout(QtWidgets.QVBoxLayout())
        self.video_loaded = False
        self.video_playing = False
        self.video_paths = video_paths
        self.current_video_index = 0
        self.top_ui = QtWidgets.QVBoxLayout()
        self.top_ui.setContentsMargins(0, 0, 0, 0)
        self.videos_navigate = QtWidgets.QHBoxLayout()
        self.previous_video_button = QtWidgets.QPushButton("Previous Video")
        self.previous_video_button.clicked.connect(self._on_click_previous_video)
        self.previous_video_button.setDisabled(True)
        self.videos_navigate.addWidget(self.previous_video_button)
        self.next_video_button = QtWidgets.QPushButton("Next Video")
        if len(self.video_paths) < 2:
            self.next_video_button.setDisabled(True)
        self.next_video_button.clicked.connect(self._on_click_next_video)
        self.videos_navigate.addWidget(self.next_video_button)
        self.filename_label = QtWidgets.QLabel()
        self.filename_label.setMargin(0)
        self.videos_navigate.setContentsMargins(0, 0, 0, 0)
        self.top_ui.addLayout(self.videos_navigate)
        self.top_ui.setContentsMargins(0, 0, 0, 0)
        self.video_viewer = QtMultimediaWidgets.QVideoWidget()
        self.player = QtMultimedia.QMediaPlayer()
        self.player.setMuted(True)
        self.trim_start = [0] * len(
            self.video_paths
        )  # the beginning point for the trim in milliseconds
        self.trim_end = [0] * len(self.video_paths)  # the end point for the trim in milliseconds
        self.play_pause = PlayPause()
        self.play_pause.fast_reverse_button.clicked.connect(self._on_click_fast_reverse)
        self.play_pause.play_pause_button.clicked.connect(self.on_click_play_pause)
        self.play_pause.fast_forward_button.clicked.connect(self._on_click_fast_forward)
        self.trim_buttons = QtWidgets.QHBoxLayout()
        self.trim_buttons.setContentsMargins(0, 0, 0, 0)
        self.trim_start_button = QtWidgets.QPushButton("Trim Start")
        self.trim_start_button.clicked.connect(self._on_click_trim_start)
        self.trim_buttons.addWidget(self.trim_start_button)
        self.trim_end_button = QtWidgets.QPushButton("Trim End")
        self.trim_end_button.clicked.connect(self._on_click_trim_end)
        self.trim_buttons.addWidget(self.trim_end_button)
        self.slider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        self.slider.setRange(0, 0)
        self.slider.sliderMoved.connect(self._on_slider_move)
        self.goto_start = GoToTime("Start Time")
        self.goto_start.goto_button.clicked.connect(self._on_goto_start)
        self.goto_end = GoToTime("End Time")
        self.goto_end.goto_button.clicked.connect(self._on_goto_end)
        self.subclip_duration_label = QtWidgets.QLabel(f"Sub-clip Duration {0}")
        self.finish_buttom = QtWidgets.QPushButton("Finish")
        self.finish_buttom.clicked.connect(self._on_click_finish)
        self.trim_control = QtWidgets.QGridLayout()
        self.trim_control.setContentsMargins(0, 0, 0, 0)
        x, y = 0, 0
        self.trim_control.addLayout(self.play_pause, 0, 0)
        self.trim_control.addLayout(self.trim_buttons, 0, 2)
        y = 0
        x += 1
        self.trim_control.addWidget(self.slider, x, y, x, -1)
        y = 0
        x += 1
        self.trim_control.addLayout(self.goto_start, x, y)
        y += 1
        self.trim_control.addLayout(self.goto_end, x, y)
        y += 1
        self.trim_control.addWidget(self.subclip_duration_label, x, y)
        y += 1
        self.trim_control.addWidget(self.finish_buttom, x, y)
        self.layout().addLayout(self.top_ui)
        self.layout().addWidget(self.video_viewer)
        self.layout().addLayout(self.trim_control)
        self.player.setVideoOutput(self.video_viewer)
        self.player.stateChanged.connect(self._state_changed)
        self.player.positionChanged.connect(self._position_changed)
        self.player.durationChanged.connect(self._duration_changed)
        self._load_video(self.video_paths[self.current_video_index])

    def _load_video(self, path):
        self.player.setMedia(QtMultimedia.QMediaContent(QtCore.QUrl.fromLocalFile(path)))
        self.filename_label.setText(os.path.split(path)[1])
        self.filename_label.resize(self.filename_label.sizeHint())
        self.trim_end[self.current_video_index] = self.player.duration()
        self._update_subclip_duration()

    def _on_click_previous_video(self, e):
        if self.current_video_index > 0:
            self.current_video_index -= 1
            self._load_video(self.video_paths[self.current_video_index])
            if self.current_video_index == 0:
                self.previous_video_button.setEnabled(False)
            if self.current_video_index < len(self.video_paths) - 1:
                self.next_video_button.setEnabled(True)

    def _on_click_next_video(self, e):
        if self.current_video_index < len(self.video_paths) - 1:
            self.current_video_index += 1
            self._load_video(self.video_paths[self.current_video_index])
            if self.current_video_index > 0:
                self.previous_video_button.setEnabled(True)
            if self.current_video_index == len(self.video_paths) - 1:
                self.next_video_button.setEnabled(False)

    def on_click_play_pause(self):
        logging.info(f"state = {self.player.state()}")
        if self.player.state() == QtMultimedia.QMediaPlayer.PlayingState:
            self.player.pause()
            logging.info("pause")
        else:
            self.player.play()
            logging.info("play")

    def _on_click_fast_forward(self):
        if self.player.state() != QtMultimedia.QMediaPlayer.StoppedState:
            self.player.setPosition(self.player.position() + 1000)

    def _on_click_fast_reverse(self):
        if self.player.state() != QtMultimedia.QMediaPlayer.StoppedState:
            self.player.setPosition(self.player.position() - 1000)

    def _on_slider_move(self, position):
        self.player.setPosition(position)

    def _on_click_trim_start(self, e):
        logging.info(f"current position {self.player.position()}")
        self.trim_start[self.current_video_index] = self.player.position()
        self.goto_start.update_label(self.trim_start[self.current_video_index] / 1000)

        if self.trim_end[self.current_video_index] < self.trim_start[self.current_video_index]:
            self.trim_end[self.current_video_index] = self.trim_start[self.current_video_index]
        self._update_subclip_duration()

    def _on_click_trim_end(self, e):
        self.trim_end[self.current_video_index] = self.player.position()
        logging.info(f"current position {self.trim_end[self.current_video_index] }")

        self.goto_end.update_label(self.trim_end[self.current_video_index] / 1000)
        if self.trim_start > self.trim_end:
            self.trim_start = self.trim_end
        self._update_subclip_duration()

    def _update_subclip_duration(self):
        subclip_duration = round(
            self.trim_end[self.current_video_index] / 1000
            - self.trim_start[self.current_video_index] / 1000,
            1,
        )
        logging.info(f"subclip_duration {subclip_duration}")
        self.subclip_duration_label.setText(f"Sub-clip Duration {subclip_duration}")

    def _on_close(self, e):
        self.loop.quit()
        self.destroy()

    def _state_changed(self, e):
        if self.player.state() == QtMultimedia.QMediaPlayer.PlayingState:
            self.play_pause.play_pause_button.setIcon(self.play_pause.pause_icon)
        else:
            self.play_pause.play_pause_button.setIcon(self.play_pause.play_icon)

    def _position_changed(self, position):
        self.slider.setValue(position)

    def _duration_changed(self, length):
        self.slider.setRange(0, length)

    def _keyPressEvent(self, e) -> None:
        key = e.key
        if key == Qt.Key_Space:
            self.on_click_play_pause()
        if key == Qt.Key_Forward:
            self._on_click_fast_forward()
        if key == Qt.Key_Back:
            self._on_click_fast_reverse()

    def _on_goto_start(self, e):
        self.player.setPosition(self.trim_start[self.current_video_index])

    def _on_goto_end(self, e):
        self.player.setPosition(self.trim_end[self.current_video_index])

    def _on_click_finish(self, e):
        args = zip(self.video_paths, self.trim_start, self.trim_end)
        with multiprocessing.Pool() as pool:
            videos = pool.imap_unordered(_trim_pool, args)
        logging.info("done")
        self.trimmed_videos = videos



from dlc_generic_analysis.viewer import ViewerWidget


class VW(ViewerWidget):
    def __init__(self):
        super(VW, self).__init__()

    def position_changed(self, position):
        super(VW, self).position_changed(position)

    def load_video(self, path: str) -> None:
        super(VW, self).load_video(path)


# def test_viewer(qtbot):
#     window = VW()
#     window.load_video("")
#     window.setBaseSize(window.sizeHint().width(), window.sizeHint().height())
#     qtbot.addWidget(window)
#     window.show()
#     window.on_play_pause()
#     window.on_fast_forward()
#     window.on_fast_reverse()
#     window.on_play_pause()



import os.path
import deeplabcut
from typing import List
from logging import info


def dlc_analyze(
    config: str,
    paths: List[str],
    gputouse: int = 0,
    filter: bool = True,
) -> (List[str], str):
    """
    DLC analyze videos or a batch of videos with the same file extension. will use the 0th paths extension for encoding
    :param config: the directory containing the config.yaml file
    :param paths: the paths to the videos to be analyzed
    :param gputouse: the CUDA capable Nvidia GPU to use for inference. If not set GPU inference will be disabled
    :param filter: if set true the data will be filtered
    :param min_confidence: value between 0 and 1 the minimum confidence for data when filtering. any data below
     min_confidence will be discarded
    :return: h5, model_slug: h5: a list of the h5 paths from the analysis, model_slug: the name of then DLC model in the
     h5 file
    """
    cfg = os.path.join(config, "config.yaml")
    _, vidtype = os.path.splitext(paths[0])
    dest, _ = os.path.split(paths[0])
    if gputouse >= 0:
        model_slug = deeplabcut.analyze_videos(
            cfg,
            paths,
            videotype=vidtype,
            destfolder=dest,
            gputouse=gputouse,
            save_as_csv=True,
            TFGPUinference=True,
        )
    else:
        model_slug = deeplabcut.analyze_videos(
            cfg,
            paths,
            videotype=vidtype,
            destfolder=dest,
            save_as_csv=True,
            TFGPUinference=False,
        )
    h5 = []
    if filter:
        deeplabcut.filterpredictions(cfg, paths, videotype=vidtype)
        for path in paths:
            h5.append(
                os.path.join(
                    dest,
                    os.path.splitext(os.path.split(path)[1])[0] + model_slug + "_filtered" + ".h5",
                )
            )
    info("analysis done")
    return h5, model_slug



from dlc_generic_analysis import MainWidget


# class TestMainWidget(MainWidget):
#     def __init__(self, title: str):
#         super(TestMainWidget, self).__init__(title=title)
#
#     def on_click_analyze(self):
#         pass
#
#     def on_click_view(self):
#         pass
#
#     def on_click_trim(self):
#         super(TestMainWidget, self).on_click_trim()
#
#
# def test_main_widget(qtbot):
#     tmw = TestMainWidget("test")
#     qtbot.addWidget(tmw)
#     tmw.on_click_analyze()
#     tmw.on_click_view()
#
#
# def test_trimmer(qtbot):
#     tmw = TestMainWidget("test")
#     qtbot.addWidget(tmw)



import os.path
from logging import info
from abc import abstractmethod
from qtpy import QtWidgets, QtCore, QtMultimedia, QtMultimediaWidgets
from qtpy.QtCore import Qt
from .gui_objects import PlayPause

try:
    from matplotlib.backends.backend_qtcairo import (
        FigureCanvasQTCairo as FigureCanvasQT,
    )
except (OSError, ImportError) as e:
    from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvasQT

    info("Cairo could not be imported. using raster AGG plots " + str(e))


class ViewerWidget(QtWidgets.QWidget):
    def __init__(self):
        """
        The viewer for your analyzed data. This window will have your videos and plots
        """
        super(ViewerWidget, self).__init__()
        self.txt = QtWidgets.QLabel("Frame #")
        self.frame_rate = 1
        self.last_frame = -1
        self.playing = False
        self.loaded = False
        self._matlab_button = QtWidgets.QPushButton("Export to MATLAB")
        self._convert_button = QtWidgets.QPushButton("Convert to MM")
        self.top_layout = QtWidgets.QHBoxLayout()
        self._play_pause = PlayPause()
        self._play_pause.play_pause_button.clicked.connect(self.on_play_pause)
        self._play_pause.fast_forward_button.clicked.connect(self.on_fast_forward)
        self._play_pause.fast_reverse_button.clicked.connect(self.on_fast_reverse)
        self._goto_button = QtWidgets.QPushButton("Go To")
        self._seek_edit = QtWidgets.QLineEdit()
        self._scrub_bar = QtWidgets.QSlider(orientation=Qt.Horizontal)
        self._scrub_bar.sliderMoved.connect(self.on_slider_move)
        self._navigate_layout = QtWidgets.QGridLayout()
        self._navigate_layout.addLayout(self._play_pause, 0, 0)
        self._navigate_layout.addWidget(self._scrub_bar)
        self.content_layout = QtWidgets.QHBoxLayout()
        self._video_player = QtMultimedia.QMediaPlayer()
        self._video_player.setMuted(True)
        self._video_player.setNotifyInterval(5)
        self._video_player.positionChanged.connect(self.position_changed)
        self._video_player.stateChanged.connect(self.state_changed)
        self._video_player.durationChanged.connect(self.duration_changed)
        self.video_viewer = QtMultimediaWidgets.QVideoWidget()
        self.content_layout.addWidget(self.video_viewer, 1)
        self._video_player.setVideoOutput(self.video_viewer)
        self.setLayout(QtWidgets.QGridLayout())
        self.layout().addLayout(self.top_layout, 0, 0)
        self.layout().addLayout(self.content_layout, 1, 0)
        self.layout().addLayout(self._navigate_layout, 2, 0)

    def state_changed(self) -> None:
        if self._video_player.state() == QtMultimedia.QMediaPlayer.PlayingState:
            info("pause_icon")
            self._play_pause.play_pause_button.setIcon(self._play_pause.pause_icon)
        else:
            info("play icon")
            self._play_pause.play_pause_button.setIcon(self._play_pause.play_icon)

    @abstractmethod
    def position_changed(self, position) -> None:
        """
        Is called every time the frame in the video changes.
        :param position: the current position in the video in milliseconds
        :return:
        """
        self._scrub_bar.setValue(position)

    def duration_changed(self, length) -> None:
        """
        called whenever the duration of a video changes IE when a new video is loaded
        :param length: the length of the video in milliseconds
        :return: None
        """
        self._scrub_bar.setRange(0, length)

    def on_play_pause(self) -> None:
        info(f"state = {self._video_player.state()}")
        if self._video_player.state() == QtMultimedia.QMediaPlayer.PlayingState:
            self.pause()
            info("pause")
        else:
            self.play()
            info("play")

    def on_fast_forward(self) -> None:
        if self._video_player.state() != QtMultimedia.QMediaPlayer.StoppedState:
            self._video_player.setPosition(self._video_player.position() + 1000)

    def on_fast_reverse(self) -> None:
        if self._video_player.state() != QtMultimedia.QMediaPlayer.StoppedState:
            self._video_player.setPosition(self._video_player.position() - 1000)

    def on_slider_move(self, position: int) -> None:
        self._video_player.setPosition(position)

    @abstractmethod
    def load_video(self, path: str) -> None:
        """
        this function is used to load new videos and load the plots
        :param path: the path to the video file to load
        :return: none
        """
        if os.path.isfile(path):
            self._video_player.setMedia(QtMultimedia.QMediaContent(QtCore.QUrl.fromLocalFile(path)))
            self.setMinimumWidth(int(self.width() / 2))

    def pause(self) -> None:
        self._video_player.pause()

    def play(self) -> None:
        self._video_player.play()

    def ms_to_frame(self, ms: int) -> float:
        return ms / 1000 * self.frame_rate

    def frame_to_ms(self, frame: int) -> float:
        return frame / self.frame_rate * 1000



import os
import numpy as np
from dlc_generic_analysis.analysis import Analysis
import pandas


class ATest(Analysis):
    def __init__(self, h5, dlc_scorer):
        Analysis.__init__(self, h5, dlc_scorer)

    def draw(self):
        pass

    def write_csv(self):
        pass


def test_analyze():
    index = pandas.MultiIndex.from_product(
        [["dlc_sc"], ["Glab", "Ment"], ["x", "y", "likelihood"]],
        names=["scorer", "bodyparts", "coords"],
    )
    data = np.array([[0, 1, 0.5, 3, 4, 0.99], [6, 7, 0.8, 1, 2, 0.6], [4, 5, 0.9, 7, 8, 0.9]])
    print(data)
    df = pandas.DataFrame(data, columns=index)
    path = "test.h5"
    df.to_hdf(path, "dlc_sc")
    at = ATest(path, "dlc_sc")
    at.draw()
    at.write_csv()
    os.remove("test.h5")



from dlc_generic_analysis import line
import numpy as np
from numpy import testing


def test_nan_line():
    testing.assert_array_equal(line.nan_line(), np.array([np.nan] * 6, dtype=np.float_))


def test_from_slope():
    test_line = line.from_slope(2, 0)
    l = np.array([2, 0, 0, 0, 1, 2], dtype=np.float_)
    testing.assert_array_equal(test_line, l)


def test_from_points_1d():
    test_line = line.from_points_1d((1, 2), (10, 20))
    np_line = np.array([2, 0, 1, 2, 10, 20], dtype=np.float_)
    print(np_line)
    testing.assert_array_equal(test_line, np_line)


def test_from_points():
    test_lines = line.from_points(np.array([[10, 1], [5, 5]]), np.array([[20, 2], [10, 0]]))
    np_lines = np.array([[0.1, 0, 10, 1, 20, 2], [-1, 10, 5, 5, 10, 0]])
    testing.assert_array_equal(np_lines, test_lines)


def test_get_slope():
    np_line = np.array([2, 0, 1, 2, 10, 20], dtype=np.float_)
    assert line.slope(np_line) == 2


def test_get_intercept():
    np_line = np.array([2, 0, 1, 2, 10, 20], dtype=np.float_)
    assert line.intercept(np_line) == 0


def test_get_end0():
    np_line = np.array([2, 0, 1, 2, 10, 20], dtype=np.float_)
    testing.assert_array_equal(line.end0(np_line), np.array([1, 2]))


def test_get_end1():
    np_line = np.array([2, 0, 1, 2, 10, 20], dtype=np.float_)
    testing.assert_array_equal(line.end1(np_line), np.array([10, 20]))



from qtpy import QtWidgets
from .trimmer import Trimmer
from abc import abstractmethod
from . import gui_utils


class MainWidget(QtWidgets.QWidget):
    def __init__(self, title: str):
        """
        The Main view of your applications where users will run analysis
        """
        super(MainWidget, self).__init__()
        self.title = title
        self.setLayout(QtWidgets.QGridLayout())
        self.title_label = QtWidgets.QLabel(self.title)
        self.layout().addWidget(self.title_label)
        self.analyze_button = QtWidgets.QPushButton("Analyze Videos")
        self.analyze_button.clicked.connect(self.on_click_analyze)
        self.layout().addWidget(self.analyze_button)
        self.view_button = QtWidgets.QPushButton("View Analyzed Video")
        self.view_button.clicked.connect(self.on_click_view)
        self.layout().addWidget(self.view_button)
        self.trim_button = QtWidgets.QPushButton("Trim Videos")
        self.trim_button.clicked.connect(self.on_click_trim)
        self.layout().addWidget(self.trim_button)

    @abstractmethod
    def on_click_analyze(self):
        """
        Runs when the user clicks the analyze button
        """
        pass

    @abstractmethod
    def on_click_view(self):
        """
        Runs when the user clicks the view button
        """
        pass

    @abstractmethod
    def on_click_trim(self):
        """
        Runs when the users clicks the trim button
        """
        vids = gui_utils.open_files(self, "Select videos to trim")
        trimmer = Trimmer(vids)
        trimmer.show()



from typing import List
from qtpy import QtWidgets


def open_files(
    parent: QtWidgets.QWidget,
    window_name: str,
    types_name: str = "Videos",
    file_types: List[str] = None,
):
    """
    creates a QfileDialog with text label for types_name file_types
    :param parent: the parent QtWidget
    :param window_name: the name of the dialog window
    :param types_name: the name of the group of file types you want to accept
    :param file_types: the file extensions you want to accept
    :return: A list of paths to files that were selected
    """
    if file_types is None:
        file_types = ["mp4", "avi", "m4v"]
    file_types_str = types_name + " ("
    for f_type in file_types:
        file_types_str += "*." + f_type + " "
    file_types_str += ");; @All Files (*)"
    files, _ = QtWidgets.QFileDialog.getOpenFileNames(
        parent,
        window_name,
        "",
        file_types_str,
        options=QtWidgets.QFileDialog.Options(),
    )
    return files


def open_dir(parent: QtWidgets.QWidget, window_name):
    """
    Opens a directory
    :param parent: the parent QWidget that launches the file dialog
    :param window_name: the name of the file dialog window
    :return: the path to the opened directory
    """
    files_dir, _ = QtWidgets.QFileDialog.getExistingDirectory(
        parent,
        window_name,
    )
    return files_dir



import os
import pandas
import abc


class Analysis(abc.ABC):
    def __init__(self, h5_path: str, dlc_scorer: str, startframe: int = 0, endframe: int = None):
        """
        Analysis:
        :param h5_path: the path of the h5 file created by analyze_videos
        :param dlc_scorer: the DeepLabCut model name
        :param startframe: the first frame of video to analyze
        :param endframe: the last frame of video to analyze
        """
        if not os.path.isfile(h5_path):
            raise FileNotFoundError(h5_path)
        self.df = pandas.read_hdf(h5_path)[dlc_scorer]
        self.start_frame = startframe
        if endframe is None:
            self.end_frame = len(self.df.index)
        else:
            self.end_frame = endframe
        assert self.end_frame > self.start_frame
        self.nframes = self.end_frame - self.start_frame
        self.outpath = os.path.split(h5_path)[0]
        self.outpath = os.path.join(h5_path, "analyzed_videos")

    @abc.abstractmethod
    def draw(self):
        """
        draw geometries from the analysis on the video and return a new video
        """
        pass

    @abc.abstractmethod
    def write_csv(self):
        """write data from the analysis to a file"""
        pass



from numba import njit
import numpy as np


nan_l = np.zeros(6)
nan_l[:] = np.nan


def nan_line():
    return nan_l.copy()


def from_slope(slope: np.ndarray, intercept: np.ndarray):
    slope = np.array(slope)
    intercept = np.array(intercept)
    if len(slope.shape) > 0:
        if slope.shape[0] >= 1:
            axis = 1
        else:
            axis = 0
    else:
        axis = 0
    return np.stack(
        [
            slope,
            intercept,
            np.zeros(slope.shape),
            intercept,
            np.ones(slope.shape),
            np.round_(intercept + slope),
        ],
        axis=axis,
    )


def from_points_1d(end0, end1):
    """
    creates a numpy array representation of a line from 2 points
    :param end0: the first point
    :param end1: the second point
    """
    slope = (end1[1] - end0[1]) / (end1[0] - end0[0])
    return np.stack(
        [
            slope,
            end0[1] - slope * end0[0],
            end0[0],
            end0[1],
            end1[0],
            end1[1],
        ]
    )


def from_points(_end0, _end1):
    """
    creates a list of lines from 2 2d lists of points
    :param _end0: a list of the first points
    :param _end1: a list of the second points
    """
    _slope = (_end1[:, 1] - _end0[:, 1]) / (_end1[:, 0] - _end0[:, 0])
    return np.stack(
        [
            _slope,
            _end0[:, 1] - _slope * _end0[:, 0],
            _end0[:, 0],
            _end0[:, 1],
            _end1[:, 0],
            _end1[:, 1],
        ],
        axis=1,
    )


@njit()
def set_ends(line: np.ndarray, points: "np.ndarray") -> "np.ndarray":
    """
    sets the ends of a line to the outermost points
    :param line:
    :param points:
    :return:
    """
    c = points[np.logical_not(np.isnan(points).any(axis=1))]
    y = c[points.shape[0] - 1, 1]
    if line[0] != 0:
        line[4] = int((y - line[1]) / line[0])
        line[5] = int(y)
    else:
        line[4] = 1
        line[5] = int(y)
    y = c[0][1]
    if line[0] != 0:
        line[2] = int((y - line[1]) / line[0])
        line[3] = int(y)
    else:
        line[2] = 1
        line[3] = int(y)
    return line


def slope(line: np.ndarray):
    return line[0]


def intercept(line: np.ndarray):
    return line[1]


def end0(line: np.ndarray):
    return line[2:4]


def end1(line: np.ndarray):
    return line[4:6]
